"""Panel Cleaner

Usage:
    pcleaner clean [<image_path> ...] [--output_dir=<output_dir>] [--profile=<profile>] [--keep-cache]
        [--save-only-mask | --save-only-cleaned] [--cache-masks] [--separate-noise-mask] [--hide-analytics]
        [--skip-text-detection] [--skip-pre-processing] [--skip-cleaning] [--skip-denoising] [--debug]
    pcleaner profile (list | new <profile_name> [<profile_path>] | add <profile_name> <profile_path> |
        open <profile_name> | delete <profile_name> | set-default <profile_name> | repair <profile_name> |
        purge-missing) [--debug]
    pcleaner ocr [<image_path> ...] [--output-path=<output_path>] [--debug]
    pcleaner config (show | open)
    pcleaner cache clear (all | models | cleaner)
    pcleaner load model [--cuda | --cpu | --both] [--force]
    pcleaner --help
    pcleaner --version

Subcommands:
    clean            Clean the given image(s). Any number of images and directories can be given.
    profile          Manage profiles. These are files storing the settings for the program.
        list         List all saved profiles.
        new          Create a new profile. Supply a name to use and, optionally, a path to save it to.
                     By default, new profiles are saved in the config directory.
        add          Add an existing profile to the list of saved profiles. This way it can be loaded by name alone.
                     Provide the name to save it under and the path the file is located at.
        open         Open a saved profile in the default editor (unless specified in the config).
        delete       Delete a saved profile.
        set-default  Set a profile as the default profile. This way it will be loaded automatically.
        repair       Repair a profile. This will remove any invalid entries and save the profile.
                     Warning: Changes to the comments won't be preserved, only settings.
        purge-missing  Remove all profiles that link to a file that doesn't exist.
    ocr              Run only the OCR on the given image(s). Any number of images and directories can be given.
                     The output will be saved in a single text file for the whole batch.
    config           View or edit the config file. This stores setting independent of profiles.
        show         Show the current configuration. This doesn't show the current profile.
        open         Open the config file in the default editor (unless specified in the config).
    cache clear      Clear the cache. This is where the program stores downloaded models and other files.
        all          Clear all cache files.
        models       Clear only the models.
        cleaner      Clear only the temporary masks and debug data.
    load models      Download the models used by the program. If neither --cuda nor --cpu is specified, the program will
                     try to use CUDA if available. If it is not, it will fall back to CPU.
                     This is done automatically when needed, but can be done manually.


Options:
    <image_path>                    One or multiple files or directories to clean.
                                    Leave blank to use the current working directory.
    -o <output_dir> --output_dir=<output_dir>  The directory to save the output files to.
                                    By default, they are saved in the same directory as the input files.
    -p <profile> --profile=<profile>  The profile to use. Specify either the name of a saved profile
                                    or a path to a profile file. If a default profile is set, you can still use
                                    the builtin default by using the name "none" or "builtin".
    -k --keep-cache                 Keep the cache files from the previous run.
                                    These are normally deleted in the text detection step.
    -m --save-only-mask             Save only the mask image. This is the overlay to clean the original image.
    -c --save-only-cleaned          Save only the cleaned image. This is the original image with the mask applied.
    -n --separate-noise-mask        Save the noise mask separately from the main mask.
    -T --skip-text-detection        Do not run the text detection AI model. This is step 1/4.
    -P --skip-pre-processing        Do not run data pre-processing. This is step 2/4.
    -C --skip-cleaning              Do not run the cleaning process. This is step 3/4.
    -D --skip-denoising             Do not run the denoising process. This an optional step 4/4.
    -s --cache-masks                Save the masks used to clean the image in the cache directory.
    -a --hide-analytics             Hide the analytics. These are the statistics about the
                                    cleaning accuracy.
    <profile_path>                  The path to the profile file to add.
    <profile_name>                  The saved name of the profile to open, delete, or set as default.
    --output-path=<output_path>     The path to save the OCR output file to.
    --cuda                          Load the torch models that support CUDA. They will only be used if supported.
    --cpu                           Load the open cv2 models that are optimized for CPU.
                                    They will only be used as a fallback, unless specified in the config.
    --both                          Load both the torch and open cv2 models.
    --force                         Force the models to be downloaded, even if they already exist.
    -d --debug                      Show debug information.
    -v --version                    Show the version and exit.
    -h --help                       Show this screen.

Examples:
    pcleaner myfolder               This will clean all images in the folder, saving the output to
                                    a folder inside myfolder called cleaned.

    pcleaner myfolder -o myoutput   This will clean all images in the folder, saving the output to
                                    a folder called myoutput, placed inside myfolder.

    pcleaner myfolder myfolder2 mypng myjpg  This will clean all images in the folders and all given files,
                                    saving the outputs to folders called cleaned, placed inside of these folders
                                    or in the parent directory of the input files, respectively.

    pcleaner myfolder -p myprofile  This will clean all the images, but use the settings from the profile
                                    called myprofile.

    pcleaner myfolder -p myprofile2 -TP  This will skip the text detection and pre-processing steps,
                                    but still run the cleaning process using the settings from the profile
                                    called myprofile2. This can be useful when you tweaked settings only
                                    related to the cleaning process. You can save time skipping the first one
                                    or two steps, since the results are saved in the cache directory (unless
                                    you choose to delete them).

"""


import time
from multiprocessing import Pool
from pathlib import Path
import itertools

from manga_ocr import MangaOcr
from docopt import magic_docopt
from logzero import logger, loglevel, DEBUG, INFO
from tqdm import tqdm
from natsort import natsorted
import torch

from pcleaner import __version__
import pcleaner.cleaner as cl
import pcleaner.config as cfg
import pcleaner.cli_utils as cli
import pcleaner.pre_processor as pp
import pcleaner.analytics as an
import pcleaner.structures as st
import pcleaner.profile_cli as pc
import pcleaner.denoiser as dn
import pcleaner.model_downloader as md


def main():

    args = magic_docopt(__doc__, version=f"Panel Cleaner {__version__}")
    # Loglevel is info by default.
    if args.debug:
        loglevel(DEBUG)
    else:
        loglevel(INFO)

    logger.debug(args)

    if args.profile:
        # Handle profile subcommands.
        config = cfg.load_config()

        if args.list:
            pc.list_profiles(config)
        elif args.new:
            pc.new_profile(config, args.profile_name, args.profile_path)
        elif args.add:
            pc.add_profile(config, args.profile_name, args.profile_path)
        elif args.open:
            pc.open_profile(config, args.profile_name)
        elif args.delete:
            pc.delete_profile(config, args.profile_name)
        elif args.set_default:
            pc.set_default_profile(config, args.profile_name)
        elif args.repair:
            pc.repair_profile(config, args.profile_name)
        elif args.purge_missing:
            pc.purge_missing_profiles(config)
        else:
            raise ValueError("Invalid profile subcommand.")

    elif args.ocr:
        config = cfg.load_config()
        run_ocr(config, args.image_path, args.output_path)

    elif args.cache and args.clear:
        config = cfg.load_config()
        clear_cache(config, args.all, args.models, args.cleaner)

    elif args.load and args.model:
        config = cfg.load_config()
        md.download_models(config, args.force, args.cuda or args.both, args.cpu or args.both)

    elif args.config:
        # Handle config subcommand.
        config = cfg.load_config()

        if args.show:
            config.show()
        elif args.open:
            cli.open_file_with_editor(cli.get_config_path(), config.profile_editor)
    else:
        # Do the actual work.
        config = cfg.load_config()
        config.load_profile(args["--profile"])
        logger.debug(config)

        if args.output_dir is None:
            args.output_dir = Path("cleaned")
            args.relative_output = True
        else:
            args.output_dir = Path(args.output_dir)

        # start timer.
        start = time.time()
        run_cleaner(
            image_paths=args.image_path,
            output_dir=args.output_dir,
            config=config,
            skip_text_detection=args.skip_text_detection,
            skip_pre_processing=args.skip_pre_processing,
            skip_cleaning=args.skip_cleaning,
            skip_denoising=args.skip_denoising,
            save_only_mask=args.save_only_mask,
            save_only_cleaned=args.save_only_cleaned,
            cache_masks=args.cache_masks,
            separate_noise_mask=args.separate_noise_mask,
            hide_analytics=args.hide_analytics,
            keep_cache=args.keep_cache,
            debug=args.debug,
        )
        # end timer.
        end = time.time()
        print(f"\nTime elapsed: {end - start:.2f} seconds")


def run_cleaner(
    image_paths: list[Path],
    output_dir: Path,
    config: cfg.Config,
    skip_text_detection: bool,
    skip_pre_processing: bool,
    skip_cleaning: bool,
    skip_denoising: bool,
    save_only_mask: bool,
    save_only_cleaned: bool,
    cache_masks: bool,
    separate_noise_mask: bool,
    hide_analytics: bool,
    keep_cache: bool,
    debug: bool,
):
    """
    Run the cleaner on the given images.

    :param image_paths: The paths to the images to clean.
    :param output_dir: The directory to save the output files to.
    :param config: The config to use.
    :param skip_text_detection: Whether to skip the text detection step.
    :param skip_pre_processing: Whether to skip the pre-processing step.
    :param skip_cleaning: Whether to skip the cleaning step.
    :param skip_denoising: Whether to skip the denoising step.
    :param save_only_mask: Whether to save only the mask.
    :param save_only_cleaned: Whether to save only the cleaned image.
    :param cache_masks: Whether to save the masks used to clean the image in the cache directory.
    :param separate_noise_mask: Whether to save the noise mask separately.
    :param hide_analytics: Whether to hide the analytics.
    :param keep_cache: Whether to keep the cache directory for the text detection step.
    :param debug: Whether to show debug information.
    """
    profile = config.current_profile

    # Catch jokesters who want to skip all 4 steps.
    if skip_text_detection and skip_pre_processing and skip_cleaning and skip_denoising:
        print("Well how about that, you want to skip all 4 steps? I guess I'm not needed here.")
        return

    cache_dir = config.get_cleaner_cache_dir()
    logger.debug(f"Cache directory: {cache_dir}")

    # If caching masks, direct the user to the cache directory.
    if cache_masks:
        print(
            f"You can find the masks being generated in real-time in the cache directory:\n\n{cache_dir}\n"
        )

    if not skip_text_detection:
        # Delete the cache directory if not explicitly keeping it.
        if len(list(cache_dir.glob("*"))) > 0 and not keep_cache:
            cli.empty_cache_dir(cache_dir)
        # Get the model file, downloading it if necessary.
        cuda = torch.cuda.is_available()
        model_path = config.get_model_path(cuda)

        print("Running text detection AI model...")
        pp.generate_mask_data(image_paths, model_path=model_path, output_dir=cache_dir)

        # Leave some extra space here if drawing analytics, so it looks better.
        if not hide_analytics:
            print("\n")

    if not skip_pre_processing:
        # Flush it so it shows up before the progress bar.
        print("Running box data Pre-Processor...", flush=True)
        # Make sure it actually flushes at all costs = wait 100 ms.
        # (It takes several seconds to load the ocr model, so this is fine.)
        time.sleep(0.1)
        if profile.pre_processor.ocr_enabled:
            mocr = MangaOcr()
        else:
            mocr = None

        ocr_analytics = []
        for json_file_path in tqdm(list(cache_dir.glob("*.json"))):
            ocr_analytic = pp.prep_json_file(
                json_file_path,
                pre_processor_conf=profile.pre_processor,
                cache_masks=cache_masks,
                mocr=mocr,
            )
            if ocr_analytic:
                ocr_analytics.append(ocr_analytic)

        if ocr_analytics and not hide_analytics:
            an.show_ocr_analytics(ocr_analytics, profile.pre_processor.ocr_max_size)

    if not skip_cleaning:
        print("Running Cleaner...")
        # Read the json files in the image directory.
        json_files = Path(cache_dir).glob("*_clean.json")

        # When denoising, we don't immediately output the cleaned image.
        # But when not, we do, since denoising is optional.
        if not skip_denoising:
            cleaner_output_dir = None
        else:
            cleaner_output_dir = output_dir

        # Zip together the json files and the out path thing.
        data = [
            st.CleanerData(
                json_file,
                cleaner_output_dir,
                cache_dir,
                profile.cleaner,
                save_only_mask,
                save_only_cleaned,
                cache_masks,
                debug,
            )
            for json_file in json_files
        ]

        cleaner_analytics_raw = []
        with Pool() as pool:
            for analytic in tqdm(pool.imap(cl.clean_page, data), total=len(data)):
                cleaner_analytics_raw.extend(analytic)

        if not hide_analytics and cleaner_analytics_raw:
            an.show_cleaner_analytics(cleaner_analytics_raw)

    if not skip_denoising:
        print("Running Denoiser...")
        # Read the json files in the image directory.
        json_files = Path(cache_dir).glob("*_mask_data.json")

        # Zip together the json files and the out path thing.
        data = [
            st.DenoiserData(
                json_file,
                output_dir,
                cache_dir,
                profile.denoiser,
                save_only_mask,
                save_only_cleaned,
                separate_noise_mask,
                cache_masks,
                debug,
            )
            for json_file in json_files
        ]

        denoise_analytics_raw = []
        with Pool() as pool:
            for analytic in tqdm(pool.imap(dn.denoise_page, data), total=len(data)):
                denoise_analytics_raw.append(analytic)

        if not hide_analytics and denoise_analytics_raw:
            an.show_denoise_analytics(
                denoise_analytics_raw,
                profile.denoiser.noise_min_standard_deviation,
                profile.cleaner.mask_max_standard_deviation,
            )

        print("Done!")


def run_ocr(config: cfg.Config, image_paths: list[Path], output_path: str | None):
    """
    Run OCR on the given images. This is a byproduct of the pre-processing step,
    expanded to all bubbles.

    :param config: The config to use.
    :param image_paths: The images to run OCR on.
    :param output_path: The path to output the results to.
    """
    cache_dir = config.get_cleaner_cache_dir()
    logger.debug(f"Cache directory: {cache_dir}")

    # Delete the cache directory if not explicitly keeping it.
    if len(list(cache_dir.glob("*"))) > 0:
        cli.empty_cache_dir(cache_dir)
    # Get the model file, downloading it if necessary.
    cuda = torch.cuda.is_available()
    model_path = config.get_model_path(cuda)

    print("Running text detection AI model...")
    pp.generate_mask_data(image_paths, model_path=model_path, output_dir=cache_dir)

    print("\n")

    # Flush it so it shows up before the progress bar.
    print("Running box data Pre-Processor...", flush=True)
    # Make sure it actually flushes at all costs = wait 100 ms.
    # (It takes several seconds to load the ocr model, so this is fine.)
    time.sleep(0.1)

    # Modify the profile to OCR all boxes.
    # Make sure OCR is enabled.
    config.current_profile.pre_processor.ocr_enabled = True
    # Make sure the max size is infinite, so no boxes are skipped in the OCR process.
    config.current_profile.pre_processor.ocr_max_size = float("inf")
    # Make sure the sus box min size is infinite, so all boxes with "unknown" language are skipped.
    config.current_profile.pre_processor.suspicious_box_min_size = float("inf")
    # Set the OCR blacklist pattern to match everything, so all text gets reported in the analytics.
    config.current_profile.pre_processor.ocr_blacklist_pattern = ".*"

    mocr = MangaOcr()
    ocr_analytics = []
    for json_file_path in tqdm(list(cache_dir.glob("*.json"))):
        ocr_analytic = pp.prep_json_file(
            json_file_path,
            pre_processor_conf=config.current_profile.pre_processor,
            cache_masks=False,
            mocr=mocr,
        )
        if ocr_analytic:
            ocr_analytics.append(ocr_analytic)

    # Output the OCRed text from the analytics.
    removed_texts = list(itertools.chain.from_iterable(a[3] for a in ocr_analytics))

    # Find and then remove the longest common prefix from the file paths.
    prefix = an.longest_common_prefix([str(Path(path).parent) for path, _ in removed_texts])
    if prefix:
        removed_texts = [(path[len(prefix) :], text) for path, text in removed_texts]
    # Remove a rogue / or \ from the start of the path, if they all have one.
    if all(path.startswith("/") or path.startswith("\\") for path, _ in removed_texts):
        removed_texts = [(path[1:], text) for path, text in removed_texts]

    removed_texts = natsorted(removed_texts, key=lambda x: x[0])

    print("\nDetected Text:")
    text = "\n".join(f"{path}: {text}" for path, text in removed_texts)
    print(text)

    if output_path is None:
        path = Path.cwd() / "detected_text.txt"
    else:
        path = Path(output_path)

    if path.exists():
        if not cli.get_confirmation(f"File {path} already exists. Overwrite?"):
            print("Aborting.")
            return
    try:
        path.write_text(text, encoding="utf-8")
        print(f"Saved detected text to {path}")
    except OSError as e:
        print(f"Failed to write detected text to {path}")
        logger.exception(e)


def clear_cache(config: cfg.Config, all_cache: bool, models: bool, images: bool):
    """
    Clear the cache.

    :param config: The config to use.
    :param all_cache: Whether to clear all cache.
    :param models: Whether to clear the model cache.
    :param images: Whether to clear the image cache.
    """
    if all_cache:
        models = True
        images = True

    if models:
        model_cache_dir = config.get_model_cache_dir()
        cli.empty_cache_dir(model_cache_dir)
        print(f"Cleared model cache at {model_cache_dir}")
        # Delete the entries in the config.
        config.default_cv2_model_path = None
        config.default_torch_model_path = None
        config.save()

    if images:
        image_cache_dir = config.get_cleaner_cache_dir()
        cli.empty_cache_dir(image_cache_dir)
        print(f"Cleared image cache at {image_cache_dir}")


if __name__ == "__main__":
    main()



import json
import re
from dataclasses import asdict
from pathlib import Path

from PIL import Image
from manga_ocr import MangaOcr

import pcleaner.ctd_interface as ctm
import pcleaner.structures as st
import pcleaner.config as cfg


def generate_mask_data(image_path: list[Path], model_path: Path, output_dir: Path) -> None:
    """
    Run the ai model to generate masks and box data for the given image,
    or all images in the given directory.

    :param image_path: Path to the image or directory of images.
    :param model_path: Path to the model file.
    :param output_dir: Path to the directory where the results will be saved.
    """

    ctm.model2annotations(model_path, image_path, output_dir)


def prep_json_file(
    json_file_path: Path,
    pre_processor_conf: cfg.PreProcessorConfig,
    cache_masks: bool,
    mocr: MangaOcr | None = None,
) -> tuple[int, tuple[int, ...], tuple[int, ...], tuple[tuple[str, str], ...]] | None:
    """
    Load the generated json file, and clean the data, leaving only the
    relevant data for the following steps.

    Check that this file wasn't already processed, and if it was, skip it.
    Processed json files have the file name ending with '_clean.json'.

    If a manga ocr object is given, run ocr on small boxes to determine whether
    they contain mere symbols, in which case these boxes do not need to be cleaned
    and can be removed from the dataset.
    The model must be initialized somewhere else to lessen coupling, and to avoid the long
    initialization time for each page.

    Analytics data is returned if the manga ocr object is given.

    :param json_file_path: Path to the json file.
    :param pre_processor_conf: Pre processor configuration, part of the profile.
    :param cache_masks: Whether to cache the masks.
    :param mocr: [Optional] Manga ocr object.
    :return: Analytics data if the manga ocr object is given, None otherwise.
    """
    # Check if the file was already processed.
    if json_file_path.name.endswith("_clean.json"):
        return
    if json_file_path.name.endswith("_mask_data.json"):
        return

    json_data = json.loads(json_file_path.read_text())

    image_path = json_data["image_path"]
    mask_path = json_data["mask_path"]
    original_path = json_data["original_path"]
    boxes = []

    for data in json_data["blk_list"]:
        # Check minimum size of box.
        x1, y1, x2, y2 = data["xyxy"]
        box_size = (x2 - x1) * (y2 - y1)
        if box_size > pre_processor_conf.box_min_size:
            # Sussy box. Discard if it's too small.
            if (
                data["language"] == "unknown"
                and box_size < pre_processor_conf.suspicious_box_min_size
            ):
                continue
            boxes.append(data["xyxy"])

    page_data = st.PageData(image_path, mask_path, original_path, boxes, [], [], [])

    # Run OCR to discard small boxes that only contain symbols.
    analytics = None
    if mocr is not None:
        page_data, analytics = ocr_check(
            page_data,
            mocr,
            pre_processor_conf.ocr_max_size,
            pre_processor_conf.ocr_blacklist_pattern,
        )

    # Pad the boxes a bit, save a copy, and then pad them some more.
    # The copy is used as a smaller mask, and the padded copy is used as a larger mask.
    page_data.grow_boxes(pre_processor_conf.box_padding_initial, st.BoxType.BOX)
    page_data.right_pad_boxes(pre_processor_conf.box_right_padding_initial, st.BoxType.BOX)

    # A shallow copy of the box list suffices, because the tuples inside are immutable.
    page_data.extended_boxes = page_data.boxes.copy()

    page_data.grow_boxes(pre_processor_conf.box_padding_extended, st.BoxType.EXTENDED_BOX)
    page_data.right_pad_boxes(
        pre_processor_conf.box_right_padding_extended, st.BoxType.EXTENDED_BOX
    )

    # Check for overlapping boxes among the extended boxes.
    # The resulting list is saved in the page_data.merged_extended_boxes attribute.
    page_data.resolve_overlaps()

    # Copy the merged extended boxes to the reference boxes and grow them once again.
    page_data.reference_boxes = page_data.merged_extended_boxes.copy()
    page_data.grow_boxes(pre_processor_conf.box_reference_padding, st.BoxType.REFERENCE_BOX)

    # Write the json file with the cleaned data.
    json_out_path = json_file_path.parent / f"{json_file_path.stem}_clean.json"
    # Remove the _image_size attribute, because it's a cached value that may be unset.
    page_data_dict = asdict(page_data)
    del page_data_dict["_image_size"]
    json_out_path.write_text(json.dumps(page_data_dict, indent=4))

    # Draw the boxes on the image and save it.
    if cache_masks:
        page_data.visualize(Path(page_data.image_path))

    return analytics


def ocr_check(
    page_data: st.PageData, mocr: MangaOcr, max_box_size: int, ocr_blacklist_pattern: str
) -> tuple[st.PageData, tuple[int, tuple[int, ...], tuple[int, ...], tuple[tuple[str, str], ...]]]:
    """
    Run OCR on small boxes to determine whether they contain mere symbols,
    in which case these boxes do not need to be cleaned and can be removed
    from the dataset.

    The page_data object is modified in place.

    Return analytics data:
    - number of boxes
    - sizes of all boxes that were ocred
    - sizes of the boxes that were removed
    - the cached file name and the text of the boxes that were removed.

    (Returning the page data isn't strictly necessary, since it's modified in place,
    but this makes that fact more explicit.)

    :param page_data: PageData object containing the data for the page.
    :param mocr: Manga ocr object.
    :param max_box_size: Maximum size of a box in pixels, to consider it for ocr.
    :param ocr_blacklist_pattern: Regex pattern to match against the ocr result.
    :return: The modified page data and Analytics data.
    """
    base_image = Image.open(page_data.image_path)
    candidate_small_bubbles = [
        box for box in page_data.boxes if page_data.box_size(box) < max_box_size
    ]
    if not candidate_small_bubbles:
        return page_data, (len(page_data.boxes), (), (), ())
    # Check if the small bubbles only contain symbols.
    # If they do, then they are probably not text.
    # Discard them in that case.
    box_sizes = []
    discarded_box_sizes = []
    discarded_box_texts: list[tuple[str, str]] = []
    for box in candidate_small_bubbles:
        cutout = base_image.crop(box)
        text = mocr(cutout)
        remove = is_not_worth_cleaning(text, ocr_blacklist_pattern)
        box_size = page_data.box_size(box)
        box_sizes.append(box_size)
        if remove:
            discarded_box_texts.append((page_data.original_path, text))
            discarded_box_sizes.append(box_size)
            page_data.boxes.remove(box)

    return page_data, (
        len(page_data.boxes) + len(discarded_box_sizes),
        tuple(box_sizes),
        tuple(discarded_box_sizes),
        tuple(discarded_box_texts),
    )


def is_not_worth_cleaning(text: str, blacklist_pattern: str) -> bool:
    """
    Check if the text is not worth cleaning.
    This is the case if the text is empty, or if it only contains symbols.
    Note that this OCR model produces full width characters.

    :param text: Text to check.
    :param blacklist_pattern: Regex pattern to match against the text.
    :return: True if the text is not worth cleaning, False otherwise.
    """
    if re.fullmatch(blacklist_pattern, text):
        return True
    return False



import os
import os.path as osp
import glob
from pathlib import Path
import cv2
import numpy as np
import json

IMG_EXT = [".bmp", ".jpg", ".png", ".jpeg"]

NP_BOOL_TYPES = (np.bool_, np.bool8)
NP_FLOAT_TYPES = (np.float_, np.float16, np.float32, np.float64)
NP_INT_TYPES = (
    np.int_,
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.uint,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
)

# https://stackoverflow.com/questions/26646362/numpy-array-is-not-json-serializable
class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.ScalarType):
            if isinstance(obj, NP_BOOL_TYPES):
                return bool(obj)
            elif isinstance(obj, NP_FLOAT_TYPES):
                return float(obj)
            elif isinstance(obj, NP_INT_TYPES):
                return int(obj)
        return json.JSONEncoder.default(self, obj)


def find_all_imgs(img_dir, abs_path=False):
    imglist = list()
    for filep in glob.glob(osp.join(img_dir, "*")):
        filename = osp.basename(filep)
        file_suffix = Path(filename).suffix
        if file_suffix.lower() not in IMG_EXT:
            continue
        if abs_path:
            imglist.append(filep)
        else:
            imglist.append(filename)
    return imglist


imread = lambda imgpath, read_type=cv2.IMREAD_COLOR: cv2.imdecode(
    np.fromfile(imgpath, dtype=np.uint8), read_type
)
# def imread(imgpath, read_type=cv2.IMREAD_COLOR):
#     img = cv2.imdecode(np.fromfile(imgpath, dtype=np.uint8), read_type)
#     return img


def imwrite(img_path, img, ext=".png"):
    suffix = Path(img_path).suffix
    if suffix != "":
        img_path = img_path.replace(suffix, ext)
    else:
        img_path += ext
    cv2.imencode(ext, img)[1].tofile(img_path)



import math
import time

import cv2
import numpy as np
import pkg_resources as pkg
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision


def scale_img(img, ratio=1.0, same_shape=False, gs=32):  # img(16,3,256,416)
    # scales img(bs,3,y,x) by ratio constrained to gs-multiple
    if ratio == 1.0:
        return img
    else:
        h, w = img.shape[2:]
        s = (int(h * ratio), int(w * ratio))  # new size
        img = F.interpolate(img, size=s, mode="bilinear", align_corners=False)  # resize
        if not same_shape:  # pad/crop img
            h, w = (math.ceil(x * ratio / gs) * gs for x in (h, w))
        return F.pad(img, [0, w - s[1], 0, h - s[0]], value=0.447)  # value = imagenet mean


def fuse_conv_and_bn(conv, bn):
    # Fuse convolution and batchnorm layers https://tehnokv.com/posts/fusing-batchnorm-and-conv/
    fusedconv = (
        nn.Conv2d(
            conv.in_channels,
            conv.out_channels,
            kernel_size=conv.kernel_size,
            stride=conv.stride,
            padding=conv.padding,
            groups=conv.groups,
            bias=True,
        )
        .requires_grad_(False)
        .to(conv.weight.device)
    )

    # prepare filters
    w_conv = conv.weight.clone().view(conv.out_channels, -1)
    w_bn = torch.diag(bn.weight.div(torch.sqrt(bn.eps + bn.running_var)))
    fusedconv.weight.copy_(torch.mm(w_bn, w_conv).view(fusedconv.weight.shape))

    # prepare spatial bias
    b_conv = (
        torch.zeros(conv.weight.size(0), device=conv.weight.device)
        if conv.bias is None
        else conv.bias
    )
    b_bn = bn.bias - bn.weight.mul(bn.running_mean).div(torch.sqrt(bn.running_var + bn.eps))
    fusedconv.bias.copy_(torch.mm(w_bn, b_conv.reshape(-1, 1)).reshape(-1) + b_bn)

    return fusedconv


def check_anchor_order(m):
    # Check anchor order against stride order for YOLOv5 Detect() module m, and correct if necessary
    a = m.anchors.prod(-1).view(-1)  # anchor area
    da = a[-1] - a[0]  # delta a
    ds = m.stride[-1] - m.stride[0]  # delta s
    if da.sign() != ds.sign():  # same order
        m.anchors[:] = m.anchors.flip(0)


def initialize_weights(model):
    for m in model.modules():
        t = type(m)
        if t is nn.Conv2d:
            pass  # nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
        elif t is nn.BatchNorm2d:
            m.eps = 1e-3
            m.momentum = 0.03
        elif t in [nn.Hardswish, nn.LeakyReLU, nn.ReLU, nn.ReLU6, nn.SiLU]:
            m.inplace = True


def make_divisible(x, divisor):
    # Returns nearest x divisible by divisor
    if isinstance(divisor, torch.Tensor):
        divisor = int(divisor.max())  # to int
    return math.ceil(x / divisor) * divisor


def intersect_dicts(da, db, exclude=()):
    # Dictionary intersection of matching keys and shapes, omitting 'exclude' keys, using da values
    return {
        k: v
        for k, v in da.items()
        if k in db and not any(x in k for x in exclude) and v.shape == db[k].shape
    }


def check_version(current="0.0.0", minimum="0.0.0", name="version ", pinned=False, hard=False):
    # Check version vs. required version
    current, minimum = (pkg.parse_version(x) for x in (current, minimum))
    result = (current == minimum) if pinned else (current >= minimum)  # bool
    if hard:  # assert min requirements met
        assert (
            result
        ), f"{name}{minimum} required by YOLOv5, but {name}{current} is currently installed"
    else:
        return result


class Colors:
    # Ultralytics color palette https://ultralytics.com/
    def __init__(self):
        # hex = matplotlib.colors.TABLEAU_COLORS.values()
        hex = (
            "FF3838",
            "FF9D97",
            "FF701F",
            "FFB21D",
            "CFD231",
            "48F90A",
            "92CC17",
            "3DDB86",
            "1A9334",
            "00D4BB",
            "2C99A8",
            "00C2FF",
            "344593",
            "6473FF",
            "0018EC",
            "8438FF",
            "520085",
            "CB38FF",
            "FF95C8",
            "FF37C7",
        )
        self.palette = [self.hex2rgb("#" + c) for c in hex]
        self.n = len(self.palette)

    def __call__(self, i, bgr=False):
        c = self.palette[int(i) % self.n]
        return (c[2], c[1], c[0]) if bgr else c

    @staticmethod
    def hex2rgb(h):  # rgb order (PIL)
        return tuple(int(h[1 + i : 1 + i + 2], 16) for i in (0, 2, 4))


def box_iou(box1, box2):
    # https://github.com/pytorch/vision/blob/master/torchvision/ops/boxes.py
    """
    Return intersection-over-union (Jaccard index) of boxes.
    Both sets of boxes are expected to be in (x1, y1, x2, y2) format.
    Arguments:
        box1 (Tensor[N, 4])
        box2 (Tensor[M, 4])
    Returns:
        iou (Tensor[N, M]): the NxM matrix containing the pairwise
            IoU values for every element in boxes1 and boxes2
    """

    def box_area(box):
        # box = 4xn
        return (box[2] - box[0]) * (box[3] - box[1])

    area1 = box_area(box1.T)
    area2 = box_area(box2.T)

    # inter(N,M) = (rb(N,M,2) - lt(N,M,2)).clamp(0).prod(2)
    inter = (
        (torch.min(box1[:, None, 2:], box2[:, 2:]) - torch.max(box1[:, None, :2], box2[:, :2]))
        .clamp(0)
        .prod(2)
    )
    return inter / (area1[:, None] + area2 - inter)  # iou = inter / (area1 + area2 - inter)


def non_max_suppression(
    prediction,
    conf_thres=0.25,
    iou_thres=0.45,
    classes=None,
    agnostic=False,
    multi_label=False,
    labels=(),
    max_det=300,
):
    """Runs Non-Maximum Suppression (NMS) on inference results

    Returns:
         list of detections, on (n,6) tensor per image [xyxy, conf, cls]
    """

    if isinstance(prediction, np.ndarray):
        prediction = torch.from_numpy(prediction)

    nc = prediction.shape[2] - 5  # number of classes
    xc = prediction[..., 4] > conf_thres  # candidates

    # Checks
    assert (
        0 <= conf_thres <= 1
    ), f"Invalid Confidence threshold {conf_thres}, valid values are between 0.0 and 1.0"
    assert 0 <= iou_thres <= 1, f"Invalid IoU {iou_thres}, valid values are between 0.0 and 1.0"

    # Settings
    min_wh, max_wh = 2, 4096  # (pixels) minimum and maximum box width and height
    max_nms = 30000  # maximum number of boxes into torchvision.ops.nms()
    time_limit = 10.0  # seconds to quit after
    redundant = True  # require redundant detections
    multi_label &= nc > 1  # multiple labels per box (adds 0.5ms/img)
    merge = False  # use merge-NMS

    t = time.time()
    output = [torch.zeros((0, 6), device=prediction.device)] * prediction.shape[0]
    for xi, x in enumerate(prediction):  # image index, image inference
        # Apply constraints
        # x[((x[..., 2:4] < min_wh) | (x[..., 2:4] > max_wh)).any(1), 4] = 0  # width-height
        x = x[xc[xi]]  # confidence

        # Cat apriori labels if autolabelling
        if labels and len(labels[xi]):
            l = labels[xi]
            v = torch.zeros((len(l), nc + 5), device=x.device)
            v[:, :4] = l[:, 1:5]  # box
            v[:, 4] = 1.0  # conf
            v[range(len(l)), l[:, 0].long() + 5] = 1.0  # cls
            x = torch.cat((x, v), 0)

        # If none remain process next image
        if not x.shape[0]:
            continue

        # Compute conf
        x[:, 5:] *= x[:, 4:5]  # conf = obj_conf * cls_conf

        # Box (center x, center y, width, height) to (x1, y1, x2, y2)
        box = xywh2xyxy(x[:, :4])

        # Detections matrix nx6 (xyxy, conf, cls)
        if multi_label:
            i, j = (x[:, 5:] > conf_thres).nonzero(as_tuple=False).T
            x = torch.cat((box[i], x[i, j + 5, None], j[:, None].float()), 1)
        else:  # best class only
            conf, j = x[:, 5:].max(1, keepdim=True)
            x = torch.cat((box, conf, j.float()), 1)[conf.view(-1) > conf_thres]

        # Filter by class
        if classes is not None:
            x = x[(x[:, 5:6] == torch.tensor(classes, device=x.device)).any(1)]

        # Apply finite constraint
        # if not torch.isfinite(x).all():
        #     x = x[torch.isfinite(x).all(1)]

        # Check shape
        n = x.shape[0]  # number of boxes
        if not n:  # no boxes
            continue
        elif n > max_nms:  # excess boxes
            x = x[x[:, 4].argsort(descending=True)[:max_nms]]  # sort by confidence

        # Batched NMS
        c = x[:, 5:6] * (0 if agnostic else max_wh)  # classes
        boxes, scores = x[:, :4] + c, x[:, 4]  # boxes (offset by class), scores
        i = torchvision.ops.nms(boxes, scores, iou_thres)  # NMS
        if i.shape[0] > max_det:  # limit detections
            i = i[:max_det]
        if merge and (1 < n < 3e3):  # Merge NMS (boxes merged using weighted mean)
            # update boxes as boxes(i,4) = weights(i,n) * boxes(n,4)
            iou = box_iou(boxes[i], boxes) > iou_thres  # iou matrix
            weights = iou * scores[None]  # box weights
            x[i, :4] = torch.mm(weights, x[:, :4]).float() / weights.sum(
                1, keepdim=True
            )  # merged boxes
            if redundant:
                i = i[iou.sum(1) > 1]  # require redundancy

        output[xi] = x[i]
        if (time.time() - t) > time_limit:
            print(f"WARNING: NMS time limit {time_limit}s exceeded")
            break  # time limit exceeded

    return output


def xywh2xyxy(x):
    # Convert nx4 boxes from [x, y, w, h] to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[:, 0] = x[:, 0] - x[:, 2] / 2  # top left x
    y[:, 1] = x[:, 1] - x[:, 3] / 2  # top left y
    y[:, 2] = x[:, 0] + x[:, 2] / 2  # bottom right x
    y[:, 3] = x[:, 1] + x[:, 3] / 2  # bottom right y
    return y


DEFAULT_LANG_LIST = ["eng", "ja"]


def draw_bbox(pred, img, lang_list=None):
    if lang_list is None:
        lang_list = DEFAULT_LANG_LIST
    lw = max(round(sum(img.shape) / 2 * 0.003), 2)  # line width
    pred = pred.astype(np.int32)
    colors = Colors()
    img = np.copy(img)
    for ii, obj in enumerate(pred):
        p1, p2 = (obj[0], obj[1]), (obj[2], obj[3])
        label = lang_list[obj[-1]] + str(ii + 1)
        cv2.rectangle(img, p1, p2, colors(obj[-1], bgr=True), lw, lineType=cv2.LINE_AA)
        t_w, t_h = cv2.getTextSize(label, 0, fontScale=lw / 3, thickness=lw)[0]
        cv2.putText(
            img,
            label,
            (p1[0], p1[1] + t_h + 2),
            0,
            lw / 3,
            colors(obj[-1], bgr=True),
            max(lw - 1, 1),
            cv2.LINE_AA,
        )
    return img



import sys
from dataclasses import dataclass, field
from pathlib import Path

import configupdater as cu
from logzero import logger

from pcleaner import cli_utils as cli
from pcleaner import model_downloader as md


RESERVED_PROFILE_NAMES = ["builtin", "none"]


@dataclass
class TextDetectorConfig:
    model_path: str | None = None

    def export_to_conf(self, config_updater: cu.ConfigUpdater) -> None:
        """
        Write the config to the config updater object.

        No add_after_section here since it is the first section.

        :param config_updater: An existing config updater object.
        """

        config_str = f"""\
        [TextDetector]
        
        # Path to the text detection model, leave empty to use the built-in model.
        # The model can be found here:
        # https://github.com/zyddnys/manga-image-translator/releases/latest
        model_path = {none_to_empty(self.model_path)}
        
        """
        config_updater.read_string(multi_left_strip(config_str))

    def import_from_conf(self, config_updater: cu.ConfigUpdater) -> None:
        """
        Read the config from the config updater object.

        :param config_updater: An existing config updater object.
        """
        if not config_updater.has_section("TextDetector"):
            logger.info("No TextDetector section found in the profile, using defaults.")
            return

        try_to_load(self, config_updater, "TextDetector", str | None, "model_path")


@dataclass
class PreProcessorConfig:
    box_min_size: int = 20 * 20
    suspicious_box_min_size: int = 200 * 200
    ocr_enabled: bool = True
    ocr_max_size: int = 30 * 100
    ocr_blacklist_pattern: str = "[～．ー！？０-９]*"
    box_padding_initial: int = 2
    box_right_padding_initial: int = 3
    box_padding_extended: int = 5
    box_right_padding_extended: int = 5
    box_reference_padding: int = 20

    def export_to_conf(self, config_updater: cu.ConfigUpdater, add_after_section: str) -> None:
        """
        Write the config to the config updater object.

        :param config_updater: An existing config updater object.
        :param add_after_section: The section to add the config after.
        """
        config_str = f"""\
        [PreProcessor]
        
        # Box sizes are given in the total number of pixels, so a box of 200x200 pixels has a size of 200 * 200 = 40000.
        # To see these boxes visualized, use the --show-masks flag when cleaning and look inside the cache folder.

        # Minimum size of any box to keep it. 
        box_min_size = {self.box_min_size}
        
        # Minimum size of a box with "unknown" language to keep it. This language is typically assigned to logos and other
        # badly behaved text elements.
        suspicious_box_min_size = {self.suspicious_box_min_size}
        
        # Whether to use OCR to detect boxes that aren't worth cleaning, like ones that only contain numbers or symbols.
        ocr_enabled = {self.ocr_enabled}
        
        # Maximum size of a box to perform OCR on.
        # These useless boxes are usually small, and OCR is slow, so use this as a cutoff.
        ocr_max_size = {self.ocr_max_size}
        
        # Regex pattern to match against OCR results. 
        # Anything matching this pattern is discarded.
        # Note that the OCR model returns full-width characters, so this pattern should match them.
        ocr_blacklist_pattern = {self.ocr_blacklist_pattern}
        
        # Padding to add to each side of a box.
        # This is added to the initial boxes created by the text detector AI.
        # These boxes are visualized in green with the --cache-masks flag.
        box_padding_initial = {self.box_padding_initial}
        
        # Padding to add to the right side of a box.
        # This extension helps to cover rubytext that floats off to the right of vertical text.
        box_right_padding_initial = {self.box_right_padding_initial}
        
        # Padding to add to each side of a box.
        # This is added to an extended set of boxes, used to cut out false positives by the text detector AI's mask.
        # These boxes are visualized in purple with the --cache-masks flag.
        box_padding_extended = {self.box_padding_extended}
        
        # Padding to add to the right side of a box.
        # This extension helps to cover rubytext that floats off to the right of vertical text.
        box_right_padding_extended = {self.box_right_padding_extended}
        
        # Padding to add to each side of a box.
        # This is added to the reference boxes used to sample the original image while analyzing what mask to use.
        # These boxes are visualized in blue with the --cache-masks flag.
        box_reference_padding = {self.box_reference_padding}

        """
        preproc_conf = cu.ConfigUpdater()
        preproc_conf.read_string(multi_left_strip(config_str))
        preproc_section = preproc_conf["PreProcessor"]
        config_updater[add_after_section].add_after.space(2).section(preproc_section.detach())

    def import_from_conf(self, config_updater: cu.ConfigUpdater) -> None:
        """
        Read the config from the config updater object.

        :param config_updater: An existing config updater object.
        """
        section = "PreProcessor"
        if not config_updater.has_section(section):
            logger.info(f"No {section} section found in the profile, using defaults.")
            return

        try_to_load(self, config_updater, section, int, "box_min_size")
        try_to_load(self, config_updater, section, int, "suspicious_box_min_size")
        try_to_load(self, config_updater, section, bool, "ocr_enabled")
        try_to_load(self, config_updater, section, int, "ocr_max_size")
        try_to_load(self, config_updater, section, str, "ocr_blacklist_pattern")
        try_to_load(self, config_updater, section, int, "box_padding_initial")
        try_to_load(self, config_updater, section, int, "box_right_padding_initial")
        try_to_load(self, config_updater, section, int, "box_padding_extended")
        try_to_load(self, config_updater, section, int, "box_right_padding_extended")
        try_to_load(self, config_updater, section, int, "box_reference_padding")


@dataclass
class CleanerConfig:
    mask_growth_step_pixels: int = 2
    mask_growth_steps: int = 11
    off_white_max_threshold: int = 240
    mask_improvement_threshold: float = 0.1
    mask_max_standard_deviation: float = 15
    debug_mask_color: tuple[int, int, int, int] = (108, 30, 240, 127)

    def export_to_conf(self, config_updater: cu.ConfigUpdater, add_after_section: str) -> None:
        """
        Write the config to the config updater object.

        :param config_updater: An existing config updater object.
        :param add_after_section: The section to add the new section after.
        """
        config_str = f"""\
        [Cleaner]
        
        # Number of pixels to grow the mask by each step. 
        # This bulks up the outline of the mask, so smaller values will be more accurate but slower.
        mask_growth_step_pixels = {self.mask_growth_step_pixels}
        
        # Number of steps to grow the mask by.
        # A higher number will make more and larger masks, ultimately limited by the reference box size.
        mask_growth_steps = {self.mask_growth_steps}
        
        # Maximum threshold for a pixel to be considered off-white.
        # The median color along the edge of a mask may not be pure white,
        # so to prevent slight off-white tones on a pure-white background,
        # anything lighter than this threshold value will be rounded up to pure white.
        off_white_max_threshold = {self.off_white_max_threshold}
        
        # Minimum improvement in standard deviation of the mask to continue shrinking it.
        # The standard deviation refers to the variation is color along the edge of a mask.
        # A low variation means that the mask sits in a solid color,
        # which means it doesn't intersect any text or other objects.
        # Setting a higher value here requires a higher improvement to consider a smaller mask,
        # to give a preference to larger masks.
        mask_improvement_threshold = {self.mask_improvement_threshold}
        
        # The maximum standard deviation of a mask to consider.
        # A high value here means a higher tolerance for the mask intersecting text or other objects,
        # which isn't a good mask, as it will require inpainting anyway.
        mask_max_standard_deviation = {self.mask_max_standard_deviation}
        
        # Color to use for the debug mask. This is a tuple of RGBA values.
        debug_mask_color = {','.join(map(str, self.debug_mask_color))}
        
        """
        cleaner_conf = cu.ConfigUpdater()
        cleaner_conf.read_string(multi_left_strip(config_str))
        cleaner_section = cleaner_conf["Cleaner"]
        config_updater[add_after_section].add_after.space(2).section(cleaner_section.detach())

    def import_from_conf(self, config_updater: cu.ConfigUpdater) -> None:
        """
        Read the config from the config updater object.

        :param config_updater: An existing config updater object.
        """
        section = "Cleaner"
        if not config_updater.has_section(section):
            logger.info(f"No {section} section found in the profile, using defaults.")
            return

        try_to_load(self, config_updater, section, int, "mask_growth_step_pixels")
        try_to_load(self, config_updater, section, int, "mask_growth_steps")
        try_to_load(self, config_updater, section, int, "off_white_max_threshold")
        try_to_load(self, config_updater, section, float, "mask_improvement_threshold")
        try_to_load(self, config_updater, section, float, "mask_max_standard_deviation")
        try:
            color_tuple: tuple[int, ...] = tuple(
                int(x) for x in config_updater["Cleaner"]["debug_mask_color"].value.split(",")
            )
            if len(color_tuple) != 4:
                raise ValueError(
                    f"Invalid color tuple length. Expected 4: 'Red, Green, Blue, Alpha' got {len(color_tuple)}"
                )
            color_tuple: tuple[int, int, int, int]
            self.debug_mask_color = color_tuple
        except (cu.NoOptionError, ValueError):
            pass


@dataclass
class DenoiserConfig:
    noise_min_standard_deviation: float = 0.25
    noise_outline_size: int = 5
    noise_fade_radius: int = 1
    colored_images: bool = False
    filter_strength: int = 10
    color_filter_strength: int = 10
    template_window_size: int = 7
    search_window_size: int = 21

    def export_to_conf(self, config_updater: cu.ConfigUpdater, add_after_section: str) -> None:
        """
        Write the config to the config updater object.

        :param config_updater: An existing config updater object.
        :param add_after_section: The section to add the new section after.
        """
        config_str = f"""\
        [Denoiser]
        
        # When a bit of text is hard to mask off, the cleaning step likely had to choose a
        # small mask, which leaves a lot of jpeg-artifacts behind, if they were around the text.
        
        # This means that the color of pixels around the edge of a mask isn't uniform,
        # which is quantified as a standard deviation. The denoiser can target masks with
        # a minimum standard deviation and denoise the area right around the mask.
        # This preserves details in the rest of the image, but removes artifacts right around where
        # the text used to be.
        
        # The minimum standard deviation of colors around the edge of a given mask
        # to perform denoising on the region around the mask.
        noise_min_standard_deviation = {self.noise_min_standard_deviation}
        
        # The thickness of an outline to denoise around a mask.
        noise_outline_size = {self.noise_outline_size}
        
        # Fade the edges of the denoised cover mask by this amount to smoothly blend
        # the denoised parts into the rest of the image.
        noise_fade_radius = {self.noise_fade_radius}
        
        # Set to True to support colored images.
        colored_images = {self.colored_images}
        
        # How strongly to denoise the image. Higher values mean more denoising.
        filter_strength = {self.filter_strength}
        
        # How strongly to denoise the color channels, as opposed to lightness.
        # Higher values mean more denoising.
        color_filter_strength = {self.color_filter_strength}
        
        # Size in pixels of the template patch that is used to compute weights. Should be odd.
        template_window_size = {self.template_window_size}
        
        # Size in pixels of the window that is used to compute weighted average for given pixel. Should be odd.
        search_window_size = {self.search_window_size}
        
        """
        denoiser_conf = cu.ConfigUpdater()
        denoiser_conf.read_string(multi_left_strip(config_str))
        denoiser_section = denoiser_conf["Denoiser"]
        config_updater[add_after_section].add_after.space(2).section(denoiser_section.detach())

    def import_from_conf(self, config_updater: cu.ConfigUpdater) -> None:
        """
        Read the config from the config updater object.

        :param config_updater: An existing config updater object.
        """
        section = "Denoiser"
        if not config_updater.has_section(section):
            logger.info(f"No {section} section found in the profile, using defaults.")
            return

        try_to_load(self, config_updater, section, float, "noise_min_standard_deviation")
        try_to_load(self, config_updater, section, int, "noise_outline_size")
        try_to_load(self, config_updater, section, int, "noise_fade_radius")
        try_to_load(self, config_updater, section, bool, "colored_images")
        try_to_load(self, config_updater, section, int, "filter_strength")
        try_to_load(self, config_updater, section, int, "color_filter_strength")
        try_to_load(self, config_updater, section, int, "template_window_size")
        try_to_load(self, config_updater, section, int, "search_window_size")


@dataclass
class Profile:
    """
    A profile is a collection of settings that can be saved and loaded from disk.
    """

    text_detector: TextDetectorConfig = TextDetectorConfig()
    pre_processor: PreProcessorConfig = PreProcessorConfig()
    cleaner: CleanerConfig = CleanerConfig()
    denoiser: DenoiserConfig = DenoiserConfig()

    def write(self, path: Path) -> bool:
        """
        Write the profile to a file.

        :param path: The path to write the profile to.
        :return: True if the profile was written successfully, False otherwise.
        """
        config_updater = cu.ConfigUpdater()
        self.text_detector.export_to_conf(config_updater)
        self.pre_processor.export_to_conf(config_updater, "TextDetector")
        self.cleaner.export_to_conf(config_updater, "PreProcessor")
        self.denoiser.export_to_conf(config_updater, "Cleaner")
        try:
            with open(path, "w") as file:
                config_updater.write(file)
            return True
        except Exception as e:
            logger.error(f"Failed to write profile to {path}:\n{e}")
            return False

    @classmethod
    def load(cls, path: Path) -> "Profile":
        """
        Load a profile from a config file.
        """
        config = cu.ConfigUpdater()
        try:
            config.read(path)
            profile = cls()
            profile.text_detector.import_from_conf(config)
            profile.pre_processor.import_from_conf(config)
            profile.cleaner.import_from_conf(config)
            profile.denoiser.import_from_conf(config)
        except Exception as e:
            logger.error(f"Failed to load profile from {path}:\n{e}")
            profile = cls()
            print("Failed to load profile, using default profile.")
        return profile


@dataclass
class Config:
    """
    The main configuration class.
    These setting are saved to disk.

    saved_profiles: Saved profiles contains a mapping from profile name to profile path, where it is saved.
    profile_editor: The program to use to edit the profile files. When blank, the default editor is used.
    """

    current_profile: Profile = Profile()
    default_profile: str | None = None
    saved_profiles: dict[str, Path] = field(default_factory=dict)
    profile_editor: str | None = None
    cache_dir: Path | None = None
    default_torch_model_path: Path | None = None  # CUDA
    default_cv2_model_path: Path | None = None  # CPU

    def show(self):
        """
        Print the current configuration to the console.
        """
        print("Current Configuration:\n")
        print(
            f"Default Profile: {self.default_profile if self.default_profile is not None else 'Built-in'}"
        )
        print("Saved Profiles:")
        if self.saved_profiles:
            for name, path in self.saved_profiles.items():
                print(f"- {name}: {path}")
        else:
            print("(No saved profiles)")
        print("")

        print(
            "Profile Editor:",
            self.profile_editor if self.profile_editor is not None else "System default",
        )
        print(
            "Cache Directory:", self.cache_dir if self.cache_dir is not None else "System default"
        )
        print(
            "Default Torch Model Path:",
            self.default_torch_model_path
            if self.default_torch_model_path is not None
            else "Not downloaded",
        )
        print(
            "Default CV2 Model Path:",
            self.default_cv2_model_path
            if self.default_cv2_model_path is not None
            else "Not downloaded",
        )

        print("\n" + "-" * 20 + "\n")
        print(f"Config file located at: {cli.get_config_path()}")
        print(f"System default cache directory: {cli.get_cache_path()}")

    def get_cache_dir(self) -> Path:
        """
        Get the cache directory.
        Get the system default cache directory if the user has not set a custom cache directory.
        """
        if self.cache_dir is None:
            return cli.get_cache_path()
        return self.cache_dir

    def get_model_cache_dir(self) -> Path:
        """
        Get the cache directory for models.
        """
        path = self.get_cache_dir() / "model"
        path.mkdir(parents=True, exist_ok=True)
        return path

    def get_cleaner_cache_dir(self) -> Path:
        """
        Get the cache directory for cleaner models.
        """
        path = self.get_cache_dir() / "cleaner"
        path.mkdir(parents=True, exist_ok=True)
        return path

    def add_profile(self, name: str, path: Path) -> None:
        """
        Add a profile to the saved profiles.
        """
        self.saved_profiles[name] = path

    def remove_profile(self, name: str) -> None:
        """
        Remove a profile from the saved profiles.
        """
        if self.default_profile == name:
            self.default_profile = None
        self.saved_profiles.pop(name, None)

    def save(self, config_path: Path | None = None) -> bool:
        """
        Write the configuration to disk.

        :param config_path: The path to write the config to. If None, the default config path is used.
        :return: True if the configuration was written successfully.
        """
        if config_path is None:
            config_path = cli.get_config_path()

        config = cu.ConfigUpdater()
        # Save the dict of saved profiles and the profile editor.
        saved_profiles_str = "\n".join(
            f"{name}={str(path.resolve())}" for name, path in self.saved_profiles.items()
        )

        conf_str = f"""\
        [General]
        
        # The default profile to use. Leave blank to use the built-in default profile.
        default_profile = {none_to_empty(self.default_profile)}
        
        # This is the command used to open the profile files.
        # If blank, the system default editor is used.
        profile_editor = {none_to_empty(self.profile_editor)}
        
        # This is the directory where the program will store temporary files.
        # If blank, the default directory is used.
        cache_dir = {none_to_empty(self.cache_dir)}
        
        # This is the path to the default torch model.
        # It is downloaded automatically if blank.
        default_torch_model_path = {none_to_empty(self.default_torch_model_path)}
        
        # This is the path to the default cv2 model.
        # It is downloaded automatically if blank.
        default_cv2_model_path = {none_to_empty(self.default_cv2_model_path)}
        
        
        [Saved Profiles]
        {saved_profiles_str}
        """
        config.read_string(multi_left_strip(conf_str))
        # Save the config.
        try:
            with open(config_path, "w", encoding="utf-8") as config_file:
                config.write(config_file)
        except OSError as e:
            logger.error(f"Failed to write configuration to {config_path}:\n{e}")
            return False
        return True

    @classmethod
    def from_config_updater(cls, conf_updater: cu.ConfigUpdater):
        """
        Create a Config object from a ConfigUpdater object.
        """

        config = cls()
        section = "General"
        if not conf_updater.has_section(section):
            print(
                f"Failed to load the {section} section from the config file. Proceeding with defaults."
            )
            return config

        try_to_load(config, conf_updater, section, str | None, "default_profile")
        try_to_load(config, conf_updater, section, str | None, "profile_editor")
        try_to_load(config, conf_updater, section, Path | None, "cache_dir")
        config.saved_profiles = {
            k: Path(v.value) for k, v in conf_updater["Saved Profiles"].items()
        }
        try_to_load(config, conf_updater, section, Path | None, "default_torch_model_path")
        try_to_load(config, conf_updater, section, Path | None, "default_cv2_model_path")

        if config.default_torch_model_path is not None:
            # Verify that the model exists and has the correct hash.
            if not config.default_torch_model_path.is_file():
                print("Configured torch model does not exist. Clearing entry.")
                config.default_torch_model_path = None
            elif not md.check_hash(config.default_torch_model_path, md.TORCH_SHA256):
                print("Configured torch model has the wrong hash. Clearing entry.")
                config.default_torch_model_path = None
        if config.default_cv2_model_path is not None:
            # Verify that the model exists and has the correct hash.
            if not config.default_cv2_model_path.is_file():
                print("Configured cv2 model does not exist. Clearing entry.")
                config.default_cv2_model_path = None
            elif not md.check_hash(config.default_cv2_model_path, md.CV2_SHA256):
                print("Configured cv2 model has the wrong hash. Clearing entry.")
                config.default_cv2_model_path = None

        return config

    def load_profile(self, profile_name: str | None = None):
        """
        Load a profile from disk, if a name is given.
        First search if the profile is saved, otherwise treat it like a path.
        For None, attempt to load the default profile.

        Special case: Reserve the name "builtin" and "none" (case insensitive) to load the built-in default profile.

        :param profile_name: Name or path of the profile to load.
        """
        # If no override is given, use the default profile.
        if profile_name is None:
            profile_name = self.default_profile

        # If the default profile is not set, use the builtin default profile.
        if profile_name is None:
            return

        # Check the reserved names.
        if profile_name.lower() in RESERVED_PROFILE_NAMES:
            return

        found_profile = cli.closest_match(profile_name, list(self.saved_profiles.keys()))
        if found_profile is not None:
            profile_path = self.saved_profiles[found_profile]
            logger.debug(f"Loading profile {found_profile} from {profile_path}")
        else:
            profile_path = Path(profile_name)
            logger.debug(f"Loading profile from {profile_path}")

        # Try to load the profile from the path.
        try:
            self.current_profile = Profile.load(profile_path)
        except Exception as e:
            logger.error(f"Failed to load profile from {profile_path}:\n{e}")
            self.current_profile = Profile()

    def get_model_path(self, cuda: bool) -> Path:
        """
        Get the path to the default model.
        Check the current profile first. If it is None or the file does not exist,
        return the default model path from the config.
        If it is None, download the model first.

        :param cuda: When true, prefer the torch model.
        """
        if self.current_profile.text_detector.model_path is not None:
            model_path = Path(self.current_profile.text_detector.model_path)
            if model_path.exists():
                return model_path
            else:
                print(
                    f"Model path {model_path} selected in your current profile does not exist. Using default model."
                )

        # Models are downloaded to the folder: cache directory/models
        cache_dir = self.get_model_cache_dir()
        if cuda:
            if self.default_torch_model_path is None:
                self.default_torch_model_path = md.download_torch_model(cache_dir)
                if self.default_torch_model_path is None:
                    print("Failed to find torch model.")
                    sys.exit(1)
                self.save()
            return self.default_torch_model_path
        else:
            if self.default_cv2_model_path is None:
                self.default_cv2_model_path = md.download_cv2_model(cache_dir)
                if self.default_cv2_model_path is None:
                    print("Failed to find cv2 model.")
                    sys.exit(1)
                self.save()
            return self.default_cv2_model_path


def load_config() -> Config:
    """
    Try to load the config from disk, otherwise create a new one.
    :return:
    """
    config_path = cli.get_config_path()

    if config_path.is_file():
        conf_updater = cu.ConfigUpdater()
        try:
            conf_updater.read(config_path)
        except OSError as e:
            logger.error(f"Failed to read configuration from {config_path}:\n{e}")
            if cli.get_confirmation("Do you want to create a new configuration file?"):
                config = Config()
                if not config.save(config_path):
                    print("\n\nProceeding with default configuration.")
                    return config
            else:
                print("Aborting.")
                sys.exit(1)
        # Parse the config from the ConfigUpdater.
        try:
            config = Config.from_config_updater(conf_updater)
        except cu.ParsingError as e:
            # When configupdater fails to parse the config, it raises a KeyError of the form:
            # ('No option `thing` found', {'key': 'thing'})
            if isinstance(e.args, tuple):
                e = e.args[0]
            logger.error(f"Failed to parse configuration from {config_path}:\n{e}")
            if cli.get_confirmation("Do you want to create a new configuration file?"):
                config = Config()
                if not config.save(config_path):
                    print("\n\nProceeding with default configuration.")
                    return config
            else:
                print("Aborting.")
                sys.exit(1)

    else:
        print("No config file found, creating a new one.")
        config = Config()
        config.save(config_path)

    return config


def try_to_load(
    obj: object, conf_updater: cu.ConfigUpdater, section: str, attr_type, attr_name: str
):
    """
    Try to load the attribute from the ConfigUpdater, validating the type.
    The key and the attribute name need to match.

    Supports attr_type:
    bool, int, float, str, Path, str | None, Path | None

    Union types with None will return none if the bare string is empty.

    :param obj: The object to assign it to.
    :param attr_name: The name of the attribute to load.
    :param section: The config section name.
    :param attr_type: The type to construct.
    :param conf_updater: The ConfigUpdater object to load from.
    """
    try:
        conf_data = conf_updater.get(section, attr_name).value
    except cu.NoOptionError as e:
        print(f"Option {attr_name} not found in {section}. Using default.")
        logger.debug(e)
        return

    # Handle bool specially.
    if attr_type == bool:
        if conf_data.lower() in ("1", "t", "true", "y", "yes", "on"):
            attr_value = True
        elif conf_data.lower() in ("0", "f", "false", "n", "no", "off"):
            attr_value = False
        else:
            print(
                f"Option {attr_name} in section {section} should be True or False, not '{conf_data}'"
            )
            return
    elif attr_type == str:
        attr_value = conf_data
    elif attr_type == str | None:
        if conf_data == "":
            attr_value = None
        else:
            attr_value = conf_data
    elif attr_type == int:
        try:
            attr_value = int(conf_data)
        except ValueError as e:
            print(
                f"Option {attr_name} in section {section} should be an integer.\n"
                f"Failed to parse '{conf_data}': {e}"
            )
            return
    elif attr_type == float:
        try:
            attr_value = float(conf_data)
        except ValueError as e:
            print(
                f"Option {attr_name} in section {section} should be floating point number.\n"
                f"Failed to parse '{conf_data}': {e}"
            )
            return
    elif attr_type == Path:
        try:
            attr_value = Path(conf_data)
        except ValueError as e:
            print(
                f"Option {attr_name} in section {section} should be file path.\n"
                f"Failed to parse '{conf_data}': {e}"
            )
            return
    elif attr_type == Path | None:
        if conf_data == "":
            attr_value = None
        else:
            try:
                attr_value = Path(conf_data)
            except ValueError as e:
                print(
                    f"Option {attr_name} in section {section} should be file path.\n"
                    f"Failed to parse '{conf_data}': {e}"
                )
                return
    else:
        logger.error(f"Unsupported attribute type {attr_type}.")
        return

    # Assign value.
    obj.__setattr__(attr_name, attr_value)


def none_to_empty(value: str | None) -> str:
    return "" if value is None else value


def multi_left_strip(string: str) -> str:
    """
    Strip all leading whitespace from all lines in a string.

    :param string: The string to strip.
    :return: The stripped string.
    """
    stripped_lines = (line.lstrip() for line in string.splitlines())
    return "\n".join(stripped_lines)



from pathlib import Path

import prettytable as pt

import pcleaner.cli_utils as cli
import pcleaner.config as cfg


def list_profiles(config: cfg.Config):
    """
    Display the dict of profiles in the config in a pretty table.

    :param config: The config object.
    """
    if not config.saved_profiles:
        print(
            "No profiles saved.\n\n"
            "Make a new profile with: (providing a path is optional)\n"
            "   pcleaner profile new <profile name> [<profile path>]\n"
            "or import a profile file with:\n"
            "   pcleaner profile add <profile name> <profile path>"
        )
        return

    def check_path(path: Path):
        if path.exists():
            return "OK"
        else:
            return "File not found"

    def check_default(profile: str | None):
        if profile == config.default_profile:
            return "Yes"
        else:
            return ""

    table = pt.PrettyTable()
    table.set_style(pt.SINGLE_BORDER)
    table.field_names = ["Profile Name", "Path", "Status", "Default"]
    # Align the path column to the left, but not the header.
    table.align["Path"] = "l"
    # Insert the builtin first.
    table.add_row(
        [
            cfg.RESERVED_PROFILE_NAMES[0],
            "",
            "OK",
            check_default(None),
        ]
    )
    for profile_name, profile_path in config.saved_profiles.items():
        table.add_row(
            [profile_name, profile_path, check_path(profile_path), check_default(profile_name)]
        )
    print(table)


def new_profile(config: cfg.Config, profile_name: str, profile_path: str | None):
    """
    Create a new profile with the given name and path.

    :param config: The config object.
    :param profile_name: The name of the profile.
    :param profile_path: The path to the profile.
    """
    if not is_valid_profile_name(config, profile_name):
        return

    # If no path is given, use the default path.
    if profile_path is None:
        profile_path = cli.get_config_path().parent / "profiles" / f"{profile_name}.conf"
    else:
        profile_path = Path(profile_path)
        # If the path is a directory, append a default filename.
        if profile_path.is_dir():
            profile_path = profile_path / f"{profile_name}.conf"

    # Check if overwriting an existing file.
    if profile_path.exists():
        if not cli.get_confirmation(f"Overwrite existing file {profile_path}?"):
            print("Aborting.")
            return

    profile_path.parent.mkdir(parents=True, exist_ok=True)
    # Create the profile file from the built-in default.
    profile = cfg.Profile()
    profile.write(profile_path)
    print(f"Profile {profile_name} created at {profile_path}.")

    # Add the profile to the config.
    config.add_profile(profile_name, profile_path)
    config.save()

    open_profile(config, profile_name)


def add_profile(config: cfg.Config, profile_name: str, profile_path: str):
    """
    Add a profile to the config.

    :param config: The config object.
    :param profile_name: The name of the profile.
    :param profile_path: The path to the profile file.
    """
    if not is_valid_profile_name(config, profile_name):
        return

    # Verify the profile path exists.
    profile_path = Path(profile_path)
    if not profile_path.is_file():
        print("Profile file not found.")
        return

    # Add the profile to the config.
    config.add_profile(profile_name, profile_path)
    config.save()

    print(f"Profile {profile_name} added.")


def open_profile(config: cfg.Config, profile_name: str):
    """
    Open the profile in the default editor, unless specified in the config.

    :param config: The config object.
    :param profile_name: The name of the profile.
    :return:
    """
    profile_path = config.saved_profiles[profile_name]
    cli.open_file_with_editor(profile_path, config.profile_editor)


def delete_profile(config: cfg.Config, profile_name: str):
    """
    Remove a profile from the config.
    Also offer to delete the profile file.

    :param config: The config object.
    :param profile_name: The name of the profile.
    """
    closest_match = cli.closest_match(profile_name, list(config.saved_profiles.keys()))

    if closest_match is None:
        print(f"Profile {profile_name} not found.")
        return

    profile_file = config.saved_profiles[closest_match]
    config.remove_profile(closest_match)
    config.save()
    print(f"Profile {closest_match} removed.\n")
    if not profile_file.is_file():
        # Already done.
        return
    if cli.get_confirmation(f"Delete profile file {profile_file}?"):
        profile_file.unlink()
        print("File deleted.")


def set_default_profile(config: cfg.Config, profile_name: str):
    """
    Set the default profile.

    :param config: The config object.
    :param profile_name: The name of the profile.
    """
    closest_match = cli.closest_match(profile_name, list(config.saved_profiles.keys()))
    if closest_match is None:
        print(f"Profile {profile_name} not found.")
        return

    config.default_profile = closest_match
    config.save()


def repair_profile(config: cfg.Config, profile_name: str):
    """
    Read as much of the profile as possible and then re-export the profile to the same path.

    :param config: The config object.
    :param profile_name: The name of the profile.
    """
    closest_match = cli.closest_match(profile_name, list(config.saved_profiles.keys()))
    if closest_match is None:
        print(f"Profile {profile_name} not found.")
        return

    profile_path = config.saved_profiles[closest_match]
    profile = cfg.Profile()
    not_salvageable = False
    try:
        profile.load(profile_path)
    except Exception as e:
        print(f"Unfixable error loading profile: {e}")
        profile = cfg.Profile()
        not_salvageable = True

    # If it wasn't salvageable, ask before overwriting.
    if not_salvageable:
        if not cli.get_confirmation(
            f"Reset profile {closest_match} to defaults? This will overwrite the file."
        ):
            print("Aborting.")
            return

    profile.write(profile_path)
    print(f"Profile {closest_match} repaired.")


def is_valid_profile_name(
    config: cfg.Config, profile_name: str, allow_reserved: bool = False
) -> bool:
    """
    Verify the profile name isn't empty or already in use.
    Normally reject reserved names, but allow them if specified.

    :param config: The config object.
    :param profile_name: The name of the profile.
    :param allow_reserved: Allow reserved names.
    :return: True if valid, False otherwise.
    """
    if not profile_name:
        print("Profile name cannot be empty.")
        return False
    if profile_name in config.saved_profiles.keys():
        print("Profile name already in use.")
        return False
    if profile_name.lower() in cfg.RESERVED_PROFILE_NAMES and not allow_reserved:
        print("Profile name is reserved.")
        return False
    return True


def purge_missing_profiles(config: cfg.Config):
    """
    Remove profiles from the config that no longer exist.

    :param config: The config object.
    """
    purged = False
    for profile_name, profile_path in list(config.saved_profiles.items()):
        if not profile_path.is_file():
            config.remove_profile(profile_name)
            purged = True
    if purged:
        config.save()
        print("Purged missing profiles.")



from PIL import Image
from logzero import logger

import pcleaner.image_ops as ops
import pcleaner.structures as st


def denoise_page(d_data: st.DenoiserData) -> st.DenoiseAnalytic:
    """
    Load the MaskData from the json file and perform the denoising process.

    :param d_data: All the data needed for the denoising process.
    :return: Analytics.
    """

    # Load all the cached data.
    mask_data = st.MaskData.from_json(d_data.json_path.read_text())
    cleaned_image = Image.open(mask_data.cleaned_path)
    mask_image = Image.open(mask_data.mask_path)

    # Alias.
    d_conf = d_data.denoiser_config

    # Filter for the min deviation to consider for denoising.
    boxes_to_denoise: list[tuple[int, int, int, int]] = [
        box
        for box, deviation in mask_data.boxes_with_deviation
        if deviation > d_conf.noise_min_standard_deviation
    ]

    noise_masks_with_coords: list[tuple[Image.Image, tuple[int, int]]] = [
        ops.generate_noise_mask(cleaned_image, mask_image, box, d_conf) for box in boxes_to_denoise
    ]

    combined_noise_mask = ops.combine_noise_masks(cleaned_image.size, noise_masks_with_coords)
    cleaned_image.paste(combined_noise_mask, (0, 0), combined_noise_mask)

    # Debug save.
    if d_data.show_masks:
        cache_out_path = d_data.cache_dir / (mask_data.original_path.stem + "_noise_mask.png")
        combined_noise_mask.save(cache_out_path)

    # Settle on the final output path for the cleaned image.
    if d_data.output_dir.is_absolute():
        final_out_path = d_data.output_dir / mask_data.original_path.name
    else:
        # Take the original image path, and place the image in a subdirectory.
        # This is for when multiple directories were passed in.
        final_out_path = (
            mask_data.original_path.parent / d_data.output_dir / mask_data.original_path.name
        )

    final_out_path.parent.mkdir(parents=True, exist_ok=True)
    final_cleaned_out_path = final_out_path.with_name(final_out_path.stem + "_clean.png")
    final_mask_out_path = final_out_path.with_name(final_out_path.stem + "_mask.png")
    final_mask_denoised_out_path = final_out_path.with_name(
        final_out_path.stem + "_denoised_mask.png"
    )

    logger.debug(f"Final output path: {final_cleaned_out_path}")

    # The arg parser should ensure that both can't be true at once, not like that'd be an issue, just plain silly.
    if not d_data.save_only_mask:
        # Save the final image.
        logger.debug(f"Saving final image to {final_cleaned_out_path}")
        cleaned_image.save(final_cleaned_out_path)

    if not d_data.save_only_cleaned:
        # Save the final image.
        if d_data.separate_noise_masks:
            logger.debug(f"Saving final mask to {final_mask_out_path}")
            mask_image.save(final_mask_out_path)

            logger.debug(f"Saving final denoised mask to {final_mask_denoised_out_path}")
            combined_noise_mask.save(final_mask_denoised_out_path)
        else:
            # Combine both the mask and the denoised mask into one image.
            combined_noise_mask.paste(mask_image, (0, 0), mask_image)
            logger.debug(f"Saving final mask to {final_mask_out_path}")
            combined_noise_mask.save(final_mask_out_path)

    # Package the analytics. We're only interested in the std deviations.
    return st.DenoiseAnalytic(tuple(deviation for _, deviation in mask_data.boxes_with_deviation))



import difflib
import os
import platform
import shutil
import subprocess
from pathlib import Path

from xdg import XDG_CONFIG_HOME, XDG_CACHE_HOME

from pcleaner import __program__


def get_config_path() -> Path:
    """
    Get the path to the configuration file for both Linux and Windows.
    """
    if platform.system() == "Linux":
        path = Path(XDG_CONFIG_HOME, __program__, __program__ + "rc")
    elif platform.system() == "Windows":
        path = Path(os.getenv("APPDATA"), __program__, __program__ + "config.ini")
    elif platform.system() == "Darwin":
        path = Path(
            os.getenv("HOME"),
            "Library",
            "Application Support",
            __program__,
            __program__ + "config.ini",
        )
    else:  # ???
        raise NotImplementedError("Your OS is currently not supported.")

    path.parent.mkdir(parents=True, exist_ok=True)
    return path


def get_cache_path() -> Path:
    """
    Get the default suggested path to the cache directory for both Linux and Windows.
    """
    if platform.system() == "Linux":
        path = Path(XDG_CACHE_HOME, __program__)
    elif platform.system() == "Windows":
        path = Path(os.getenv("APPDATA"), __program__, "cache")
    elif platform.system() == "Darwin":
        path = Path(
            os.getenv("HOME"),
            "Library",
            "Caches",
            __program__,
        )
    else:  # ???
        raise NotImplementedError("Your OS is currently not supported.")

    path.mkdir(parents=True, exist_ok=True)
    return path


def open_file_with_editor(path: Path, configured_opener: str | None) -> None:
    """
    Open the given file with the given editor.
    If no editor is given, use the default editor for the current OS.

    :param path: The path to the file to open.
    :param configured_opener: The configured editor to use.
    """
    if platform.system() == "Linux":
        opener = "xdg-open" if configured_opener is None else configured_opener
        print(f"Opening {path} with {opener}.")
        # Run and detatch the process, so this one can exit.
        subprocess.Popen([opener, path], start_new_session=True)

        print("If nothing happens, try setting the profile_editor option in the config.")

    elif platform.system() == "Windows":
        opener = "start" if configured_opener is None else configured_opener
        print(f"Opening {path} with {opener}.")
        os.system(f"{opener} {path}")

        print("\nIf nothing happens, try setting the 'profile_editor' option in the config.")
    elif platform.system() == "Darwin":
        opener = "open" if configured_opener is None else configured_opener
        print(f"Opening {path} with {opener}.")
        os.system(f"{opener} {path}")

        print("\nIf nothing happens, try setting the 'profile_editor' option in the config.")
    else:
        print(
            f"Your OS ({platform.system()}) is not supported, but you can open the path manually:\n"
            f"Profile path: {path}"
        )


def ask_then_delete(target_dir) -> bool:
    """
    Ask to delete a directory.
    On Linux, check if trash-cli is installed and offer to trash the item instead.
    If that fails, ask to delete it.

    :param target_dir:
    :return: True if the directory was deleted.
    """

    if shutil.which("trash"):
        if get_confirmation(f"{target_dir} already exists. Move to trash?", default=True):
            if subprocess.run(["trash-put", str(target_dir)]).returncode == 0:
                return True
            else:
                print("Failed to move to trash.")
        else:
            return False

    if get_confirmation(f"{target_dir} already exists. Do you want to delete it?", default=False):
        shutil.rmtree(target_dir)

    # Perform a final check to see if the directory was deleted.
    return not target_dir.exists()


def get_confirmation(prompt: str, default: bool | None = None) -> bool:
    """
    Get confirmation from the user.
    Repeat if the input wasn't valid.

    :param prompt: The prompt to display to the user.
    :param default: The default value to return if the user doesn't input anything.
    :return: The user's confirmation.
    """
    match default:
        case None:
            prompt += " [y/n]"
        case True:
            prompt += " [Y/n]"
        case False:
            prompt += " [y/N]"

    prompt += " > "

    while True:
        user_input = input(prompt)
        if user_input.lower().strip().startswith("y"):
            return True
        elif user_input.lower().strip().startswith("n"):
            return False
        elif user_input == "" and default is not None:
            return default
        else:
            print("Invalid input. Please try again.")


def empty_cache_dir(cache_dir: Path) -> None:
    """
    Empty the cache directory.
    Only attempt to delete .png images and .json files.
    Or .pt and .onnx files for PyTorch and ONNX models.
    This limits the damage if this points to the wrong directory.
    """
    for item in cache_dir.iterdir():
        if item.suffix in [".png", ".json", ".pt", ".onnx"]:
            item.unlink()


def closest_match(word: str, choices: list[str]) -> str | None:
    """
    Return the closest match for the given word in the list of choices.
    If no good match is found, return None.
    """
    if word in choices:
        return word
    else:
        # Find the closest match using difflib:
        closest = difflib.get_close_matches(word, choices, 1, 0.5)  # 0.6 is the default threshold
        if closest:
            return str(closest[0])
        else:
            return None



import itertools
import shutil
from collections import defaultdict
from pathlib import Path

import colorama as clr
from natsort import natsorted

import pcleaner.structures as st


def terminal_width() -> int:
    # Use a fallback of 50 for unsupported terminals.
    width = shutil.get_terminal_size((50, 50)).columns
    # Make sure the width is at least 20, to avoid weirdness.
    return max(width, 20)


def show_ocr_analytics(
    analytics: list[tuple[int, tuple[int, ...], tuple[int, ...], tuple[tuple[str, str], ...]]],
    max_ocr_size: int,
) -> None:
    """
    The analytics are gathered from the OCR pre-processing which has the goal of
    finding small boxes and to check if they contain text worth cleaning.

    The analytics for each page in the list consist of the following:
    - The number of boxes on the page.
    - The sizes of the small boxes.
    - The sizes of the boxes that were removed.
    - The original file path and the removed text.

    Various statistics are then shown.

    :param analytics: The analytics gathered from each pre-processed page.
    :param max_ocr_size: The maximum size of a box to be considered small.
    """
    num_boxes = sum(a[0] for a in analytics)
    num_small_boxes = sum(len(a[1]) for a in analytics)
    small_box_sizes = list(itertools.chain.from_iterable(a[1] for a in analytics))
    removed_box_sizes = list(itertools.chain.from_iterable(a[2] for a in analytics))
    small_box_ratio = f"{num_small_boxes / num_boxes:.0%}" if num_boxes > 0 else "N/A"
    removed_box_ratio = f"{len(removed_box_sizes) / num_boxes:.0%}" if num_boxes > 0 else "N/A"
    removed_among_small_ratio = (
        f"{len(removed_box_sizes) / num_small_boxes:.0%}" if num_small_boxes > 0 else "N/A"
    )
    # Partition the small box sizes into 5 rough size ranges, equally spaced.
    part_size = 500 if max_ocr_size < 500 * 10 else max_ocr_size // 10
    partitioned_small_box_sizes = partition_list(
        small_box_sizes, partition_size=part_size, max_value=max_ocr_size
    )
    partitioned_removed_boxes = partition_list(
        removed_box_sizes, partition_size=part_size, max_value=max_ocr_size
    )
    print("\nOCR Analytics")
    print("-------------")
    print(
        f"Number of boxes: {num_boxes} | "
        f"Number of small boxes: {num_small_boxes} ({small_box_ratio})"
    )
    print(
        f"Number of removed boxes: {len(removed_box_sizes)} ({removed_box_ratio} total, {removed_among_small_ratio} of small boxes)"
    )
    if small_box_sizes:
        print("\nSmall box sizes:")
        draw_pretty_ocr_result_chart(partitioned_small_box_sizes, partitioned_removed_boxes)
    else:
        print("No not-removed small boxes found.\n")

    # Show removed texts.
    removed_texts = list(itertools.chain.from_iterable(a[3] for a in analytics))

    # Find and then remove the longest common prefix from the file paths.
    prefix = longest_common_prefix([str(Path(path).parent) for path, _ in removed_texts])
    if prefix:
        removed_texts = [(path[len(prefix) :], text) for path, text in removed_texts]
    # Remove a rogue / or \ from the start of the path, if they all have one.
    if all(path.startswith("/") or path.startswith("\\") for path, _ in removed_texts):
        removed_texts = [(path[1:], text) for path, text in removed_texts]

    removed_texts = natsorted(removed_texts, key=lambda x: x[0])

    print("\nRemoved bubbles:")
    for path, text in removed_texts:
        print(f"Page {path}: {text}")

    print("\n")


def draw_pretty_ocr_result_chart(
    small_box_data: list[tuple[str, int]], removed_box_data: list[tuple[str, int]]
) -> None:
    """
    Draw a pretty stacked bar chart.

    :param small_box_data: The data for the small boxes.
    :param removed_box_data: The data for the removed boxes.
    """
    width = terminal_width()
    data_array = [
        (label, lower, upper)
        for (label, lower), (_, upper) in zip(small_box_data, removed_box_data)
    ]
    max_label_width = max(len(label) for label, _, _ in data_array)
    bar_width = width - max_label_width - 16  # 16 for the spacing and trailing number and padding.
    max_value = max(lower + upper for _, lower, upper in data_array)

    for label, lower, upper in data_array:
        lower_bar = "█" * int((lower - upper) / max_value * bar_width)
        upper_bar = "█" * int(upper / max_value * bar_width)
        print(
            f"{label:<{max_label_width}}: "
            f"{lower_bar}{clr.Fore.RED}{upper_bar} {upper}{clr.Style.RESET_ALL} / {lower}"
        )

    # Show legend.
    print(f"\n█ Small boxes | {clr.Fore.RED}█ Removed boxes{clr.Style.RESET_ALL}")


def draw_cleaner_histogram(data: dict[str, tuple[int, int]]) -> None:
    """
    Draw a histogram of the cleaner results.

    :param data: A dict of the mask name, (perfect uses, all uses).
    """

    width = terminal_width()
    # Find the maximum value in the data. (at least 1 to avoid division by zero)
    max_value = max(list(map(lambda x: x[1], data.values())))
    max_value = max(max_value, 1)

    # Longest key.
    max_label_width = max(len(k) for k in data.keys())
    bar_width = width - max_label_width - 16  # 16 for the spacing and trailing number and padding.

    # Draw the histogram.
    for key, value in data.items():
        # Use a Unicode character for the bar.
        left_bar = "█" * int(value[0] / max_value * bar_width)
        right_bar = "█" * int((value[1] - value[0]) / max_value * bar_width)
        # If both are 0 in length, but one of them isn't 0 in value, draw the right-most one as a sliver.
        if not left_bar and not right_bar and value[1] > value[0]:
            right_bar = "▏"
        elif not left_bar and not right_bar and value[0] > 0:
            left_bar = "▏"

        print(
            f"{key:<{max_label_width}}: "
            f"{clr.Fore.CYAN}{left_bar}{clr.Style.RESET_ALL}{right_bar} "
            f"{clr.Fore.CYAN}{value[0]}{clr.Fore.RESET} / {value[1]}"
        )

    # Show legend.
    print(f"\n{clr.Fore.CYAN}█ Perfect{clr.Style.RESET_ALL} | █ Total")


def partition_list(
    data: list[int],
    *,
    partition_size: int = 500,
    min_value: int | None = 0,
    max_value: int | None = None,
) -> list[tuple[str, int]]:
    """
    Partition the list into tuples of the form (range, count).
    The ranges are saved as a string: "0-499", "500-999", etc.
    If no max_value is given, the maximum value in the data is used.
    If min_value is None, the minimum value in the data is used. Otherwise it defaults to 0.

    :param data: The data to partition.
    :param partition_size: [Optional] The size of the partitions. Default: 500.
    :param min_value: [Optional] The minimum value to partition. Default: 0.
    :param max_value: [Optional] The maximum value to partition to. Default: None.
    :return: The partitioned data as a list of tuples.
    """
    if max_value is None:
        max_value = max(data)

    max_label_len = len(str(max_value))

    partition = [
        (
            f"{i * partition_size: >{max_label_len}}-{(i + 1) * partition_size: >{max_label_len}}",
            len([d for d in data if i * partition_size <= d < (i + 1) * partition_size]),
        )
        for i in range(min_value, max_value // partition_size)
    ]

    return partition


def show_cleaner_analytics(analytics: list[tuple[Path, bool, int, float]]):
    """
    Present the analytics gathered from the cleaning process.

    The analytics for each page in the list consist of the following tuple:
    list[original_image_path, mask_found, mask_index, border_deviation]

    :param analytics: The analytics gathered from each page.
    """

    total_boxes = len(analytics)
    masks_succeeded = sum(1 for analytics in analytics if analytics[1])
    perfect_masks = sum(1 for analytics in analytics if analytics[1] and analytics[3] == 0)
    average_border_deviation = (
        f"{sum(analytics[3] for analytics in analytics if analytics[1]) / masks_succeeded:.2f}"
        if masks_succeeded
        else "N/A"
    )
    success_rate = f"{masks_succeeded / total_boxes:.0%}" if total_boxes else "N/A"
    perfect_mask_rate = f"{perfect_masks / total_boxes:.0%}" if total_boxes else "N/A"
    masks_failed = total_boxes - masks_succeeded

    highest_mask_index = max(analytics[2] for analytics in analytics if analytics[1])
    # Count the number of times each mask index was used.
    mask_usages_by_index = [0] * (highest_mask_index + 1)
    # Count the number of times each mask was perfect.
    perfect_mask_usages_by_index = [0] * (highest_mask_index + 1)
    for analytic in analytics:
        if analytic[1]:
            mask_usages_by_index[analytic[2]] += 1
            if analytic[3] == 0:
                perfect_mask_usages_by_index[analytic[2]] += 1

    # Print the analytics.
    print("\nMask Fitment Analytics")
    print("----------------------")
    print(
        f"Total boxes: {total_boxes} | "
        f"Masks succeeded: {masks_succeeded} ({success_rate}) | "
        f"Masks failed: {clr.Fore.RED}{masks_failed}{clr.Fore.RESET}"
    )
    print(
        f"Perfect masks: {clr.Fore.CYAN}{perfect_masks}{clr.Fore.RESET} ({perfect_mask_rate}) | "
        f"Average border deviation: {average_border_deviation}"
    )
    print("\nMask usage by mask size (smallest to largest):")
    # Make a dict of {index: list contents} for the mask usages, except the last key is "Box mask".
    mask_usages_dict = {
        f"Mask {index}": perfect_total
        for index, perfect_total in enumerate(
            zip(perfect_mask_usages_by_index, mask_usages_by_index)
        )
        if index < len(mask_usages_by_index) - 1
    }
    mask_usages_dict["Box mask"] = perfect_mask_usages_by_index[-1], mask_usages_by_index[-1]

    draw_cleaner_histogram(mask_usages_dict)

    # Find out what pages had how many failures.
    # Structure: {page_path: {"succeeded": 0, "failed": 0}}
    pages_with_success_and_fails_dict: defaultdict[Path, defaultdict[str, int]] = defaultdict(
        lambda: defaultdict(int)
    )
    for analytics in analytics:
        pages_with_success_and_fails_dict[analytics[0]][
            "succeeded" if analytics[1] else "failed"
        ] += 1

    if not pages_with_success_and_fails_dict:
        print("All bubbles were successfully masked.")
        return

    # Flatten the dict into a list of tuples: (page_path, succeeded, failed).
    # Only include pages with at least one failure.
    pages_with_success_and_fails: list[tuple[Path, int, int]] = [
        (page_path, counts["succeeded"], counts["failed"])
        for page_path, counts in pages_with_success_and_fails_dict.items()
        if counts["failed"]
    ]

    # Find and then remove the longest common prefix from the file paths.
    prefix = longest_common_prefix(
        [str(path.parent) for path, _, _ in pages_with_success_and_fails]
    )
    if prefix:
        pages_with_success_and_fails = [
            (path.relative_to(prefix), succeeded, failed)
            for path, succeeded, failed in pages_with_success_and_fails
        ]

    # Sort them list by file path.
    pages_with_success_and_fails = natsorted(pages_with_success_and_fails, key=lambda x: x[0])

    print("\nPages with failures / total:")
    for page_path, succeeded, failed in pages_with_success_and_fails:
        print(f"{page_path}: {clr.Fore.RED}{failed}{clr.Fore.RESET} / {succeeded+failed}")

    print("\n")


def show_denoise_analytics(
    analytics: list[st.DenoiseAnalytic],
    configured_deviation_threshold: float,
    max_deviation_threshold: float,
):
    """
    Present the analytics gathered from the denoising process.

    :param analytics: The analytics gathered from each page, here standard deviations.
    :param configured_deviation_threshold: The configured deviation threshold to de denoising.
    :param max_deviation_threshold: The maximum deviation threshold to discard the mask.
    """
    print("\nDenoising Analytics")
    print("-------------------")

    # Get the standard deviations into one single list.
    std_deviations: list[float] = [
        *itertools.chain.from_iterable(analytic.std_deviations for analytic in analytics)
    ]

    # Split the standard deviations into two lists: one for the ones that were below the threshold,
    # and one for the ones that were above the threshold.
    below_threshold = [
        std_dev for std_dev in std_deviations if std_dev <= configured_deviation_threshold
    ]
    above_threshold = [
        std_dev for std_dev in std_deviations if std_dev > configured_deviation_threshold
    ]

    total_masks = len(std_deviations)
    above_threshold_count = len(above_threshold)
    above_ratio = f"{above_threshold_count / total_masks:.0%}" if total_masks else "N/A"

    print(
        f"Total masks: {total_masks} | Masks denoised: {clr.Fore.MAGENTA}{above_threshold_count}{clr.Fore.RESET} "
        f"({above_ratio})"
    )
    print(
        f"Minimum deviation to denoise: {configured_deviation_threshold} | "
        f"Maximum allowed deviation: {max_deviation_threshold}"
    )
    print(f"Standard deviation around masks:\n")

    # Show the lists in a graph.
    draw_denoise_histogram(
        below_threshold,
        above_threshold,
        max_deviation_threshold,
    )
    print("")


def draw_denoise_histogram(
    below_threshold: list[float],
    above_threshold: list[float],
    max_value: float,
):
    """
    Draw a histogram of the deviations.

    :param below_threshold: The standard deviations that were below the threshold.
    :param above_threshold: The standard deviations that were above the threshold.
    :param max_value: The maximum deviation threshold to discard the mask.
    """
    # Partition the values into 10 buckets.
    # Use a cubic polynomial that passes through 0 and the max value.
    # The polynomial is a simple f(x) = ax^3, meaning the coefficient a
    # must be max_value/10^3.
    exponent = 3
    cubic_coefficient = max_value / 10**exponent
    # The buckets contain: (min, max)
    bucket_ranges = (
        (cubic_coefficient * (i - 1) ** exponent, cubic_coefficient * i**exponent)
        for i in range(1, 11)
    )
    # Set the precision so that the second bucket isn't 0.
    # Basically, if the value is 0.00143, we want a precision of 3 to still see that 1.
    second_bucket_min = str(cubic_coefficient * 1**exponent).split(".")[1]
    # Count the zeroes from the left, and add 1 to only show up to the first non-zero digit.
    precision = len(second_bucket_min) - len(second_bucket_min.lstrip("0")) + 1
    # The buckets contain: (min, max, count_below_threshold, count_above_threshold)
    buckets: list[tuple[str, str, int, int]] = [
        (
            f"{b_min:.{precision}f}",
            f"{b_max:.{precision}f}",
            sum(1 for std_dev in below_threshold if b_min <= std_dev < b_max),
            sum(1 for std_dev in above_threshold if b_min <= std_dev < b_max),
        )
        for b_min, b_max in bucket_ranges
    ]
    # Format the ranges to look better.
    number_width = len(str(f"{max_value:.{precision}f}"))
    buckets_labeled: list[tuple[str, int, int]] = [
        (
            f"{b_min: >{number_width}}-{b_max: >{number_width}}",
            below,
            above,
        )
        for b_min, b_max, below, above in buckets
    ]
    # Get the maximum count in the buckets to rescale them.
    max_count = max((below + above for _, below, above in buckets_labeled))
    max_count = max(max_count, 1)
    # Rescale the buckets.

    width = terminal_width()
    max_label_length = max(len(label) for label, _, _ in buckets_labeled)
    bar_width = width - max_label_length - 16  # 16 is for the spaces and the brackets.

    buckets_labeled = [
        (b_range, int(below * bar_width / max_count), int(above * bar_width / max_count))
        for b_range, below, above in buckets_labeled
    ]

    # Draw the buckets.
    for b_range, below, above in buckets_labeled:
        print(
            f"{b_range} : {'█' * below}{clr.Fore.MAGENTA}{'█' * above} {above}{clr.Fore.RESET} / {below+above}"
        )

    # Draw the legend.
    print(f"\n{clr.Fore.MAGENTA}█ Denoised{clr.Fore.RESET} | █ Total")


def longest_common_prefix(strings: list[str]) -> str:
    """
    Finds the longest common prefix for a list of strings.

    :param strings: The list of strings to find the longest common prefix for.
    :return: The longest common prefix.
    """
    if not strings:
        return ""

    prefix = strings[0]
    for string in strings[1:]:
        while string.find(prefix) != 0:
            prefix = prefix[:-1]
            if not prefix:
                return ""
    return prefix
